package gr.ilsp.fmc.aligner.factory;

import gr.ilsp.fmc.exporter.XSLTransformer;
import gr.ilsp.fmc.utils.BitextUtils;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.FilenameUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import com.google.common.collect.Lists;

/**
 * @author prokopis
 *
 */
public abstract class Aligner {
  
	private static final Logger logger = LoggerFactory.getLogger(Aligner.class);
	protected Properties properties ;
	protected String sourceLang ;
	protected String targetLang ;
	private static final String BOILERPLATE = "boilerplate";
	private static final String CRAWLINFO = "crawlinfo";
	protected static final String PARA_NAME = "p";
	protected boolean useBoilerplateParagraphs = false;
	protected SentenceSplitter sentenceSplitter = new SentenceSplitter();

	public abstract AlignmentStats process(File sourceFile, File targetFile, File tmxFile) throws IOException, Exception;
    public abstract void destroy ();

	public void processCesAlignList(File cesAlignList, File outputTMXList, File outputHTMLTMXList) throws IOException, Exception {
    	logger.info("Aligning sentences in document pairs using "+ this.getClass().getSimpleName());
    	logger.debug("... from cesAlign list " + cesAlignList);
		List<String> lines = FileUtils.readLines(cesAlignList);
        List<File> tmxFiles = new ArrayList<File>();
        List<File> htmlTmxFiles = new ArrayList<File>();
        List<Integer> alignmentsPerFile = new ArrayList<Integer>();
        Map<String, String> basenameMap = new HashMap<String, String>();
        Map<String, Integer> alignmentsMap = new HashMap<String, Integer>();
        Map<String, Integer> sourceSentsMap = new HashMap<String, Integer>();
        Map<String, Integer> targetSentsMap = new HashMap<String, Integer>();
        XSLTransformer xslTransformer = new XSLTransformer("http://nlp.ilsp.gr/xslt/ilsp-fc/tmx2html.xsl");

        for (String line : lines) {
            File cesAlignFile = new File(line.trim());
        	logger.debug("cesAlign file " + cesAlignFile.getAbsolutePath());
            String cesAlignBasename = FilenameUtils.getBaseName(cesAlignFile.getAbsolutePath());
            DocumentBuilderFactory dbFactory = DocumentBuilderFactory.newInstance();
            DocumentBuilder dBuilder = dbFactory.newDocumentBuilder();

            try {
                Document doc = dBuilder.parse(cesAlignFile);
                NodeList translationNodes = doc.getDocumentElement().getElementsByTagName("translation");
                for(int i=0; i< translationNodes.getLength(); i++){
                    Element translationElement = (Element)translationNodes.item(i);
                    File transFile = new File(translationElement.getAttribute("trans.loc"));
                    basenameMap.put(translationElement.getAttribute("xml:lang"), FilenameUtils.getBaseName(transFile.getName()));
                }

    			File sourceFile = new File(FilenameUtils.concat(cesAlignFile.getParent(), basenameMap.get(sourceLang) + ".xml"));;
    			File targetFile = new File(FilenameUtils.concat(cesAlignFile.getParent(), basenameMap.get(targetLang) + ".xml"));;
    			File tmxFile = 	new File(FilenameUtils.concat(cesAlignFile.getParent(), cesAlignBasename + ".tmx")); 

    			AlignmentStats alignmentStats = this.process(sourceFile, targetFile, tmxFile);
    			tmxFiles.add(tmxFile);

    			File htmlTmxFile =   new File(FilenameUtils.removeExtension(tmxFile.getAbsolutePath()) + ".tmx.html");
    	        //IOtools.tmxTOhtml(tmxFile.getAbsolutePath(), htmlTmxFile.getAbsolutePath());
    	        xslTransformer.setBaseDir(tmxFile.getParent());
    	        xslTransformer.transform(tmxFile, htmlTmxFile);
    	        
    	        htmlTmxFiles.add(htmlTmxFile);
    			basenameMap.clear();

    			alignmentsPerFile.add(alignmentStats.getAlignmentsSize());
    			String documentAlignmentMethod = BitextUtils.getDocumentAlignmentMethod(cesAlignFile);
    			int count;
    			
    			count = sourceSentsMap.containsKey(documentAlignmentMethod) ? sourceSentsMap.get(documentAlignmentMethod) : 0;
    			sourceSentsMap.put(documentAlignmentMethod, count + alignmentStats.getSourceSentencesSize());

    			count = targetSentsMap.containsKey(documentAlignmentMethod) ? targetSentsMap.get(documentAlignmentMethod) : 0;
    			targetSentsMap.put(documentAlignmentMethod, count + alignmentStats.getTargetSentencesSize());

    			count = alignmentsMap.containsKey(documentAlignmentMethod) ? alignmentsMap.get(documentAlignmentMethod) : 0;
    			alignmentsMap.put(documentAlignmentMethod, count + alignmentStats.getAlignmentsSize());
    			
    			logger.debug("Exported results to tmx file " + tmxFile);
    			logger.debug("Exported results to html tmx file " + htmlTmxFile);

            } catch (Exception ex) {
            	logger.warn(ex.getMessage());
            }
        }
        
        generateTmxListFiles(outputTMXList, outputHTMLTMXList, lines, tmxFiles,
				htmlTmxFiles, alignmentsPerFile, alignmentsMap, sourceSentsMap,
				targetSentsMap);    			
	}

	private void generateTmxListFiles(File outputTMXList,
			File outputHTMLTMXList, List<String> lines, List<File> tmxFiles,
			List<File> htmlTmxFiles, List<Integer> alignmentsPerFile,
			Map<String, Integer> alignmentsMap,
			Map<String, Integer> sourceSentsMap,
			Map<String, Integer> targetSentsMap) throws FileNotFoundException,
			UnsupportedEncodingException {
		StringBuffer htmlTmxContent = new StringBuffer();
        htmlTmxContent.append("<html xmlns=\"http://www.w3.org/1999/xhtml\">");
        for (File htmlTmxFile : htmlTmxFiles){
        	htmlTmxContent.append("<br />"+"<a href=\""+htmlTmxFile.getAbsoluteFile().toURI()+
        			"\">\n"+htmlTmxFile.getAbsoluteFile().toURI()+"</a>"+"\n");
        }
        htmlTmxContent.append("</html>\n");

        PrintWriter htmlTmxWriter = new PrintWriter(outputHTMLTMXList, "UTF-8");
		htmlTmxWriter.print(htmlTmxContent.toString());
		htmlTmxWriter.close();
		logger.info("Created list of html tmx results in " + outputHTMLTMXList.getAbsolutePath());    			

        StringBuffer tmxContent = new StringBuffer();
        int tmxI = 0;
        for (File tmxFile : tmxFiles){
        	tmxContent.append(tmxFile.getAbsoluteFile());
        	tmxContent.append(" :: " + alignmentsPerFile.get(tmxI++) + " alignments");
        	tmxContent.append("\n");	
        }
        tmxContent.append("\n"
        		+ "Document pairs processed by aligner: " + tmxFiles.size()
        		+ " out of " + lines.size());
        tmxContent.append("\n");
        
        tmxContent.append("\n"
        		+ "Alignments per type:\n" ) ;
        for (String key: Lists.newArrayList(alignmentsMap.keySet())) {
            tmxContent.append("\t" + key + ":" + alignmentsMap.get(key) +"\n");        	
        }
        tmxContent.append("\n");
        tmxContent.append("\n"
        		+ "SL sentences per type:\n" ) ;
        for (String key: Lists.newArrayList(sourceSentsMap.keySet())) {
            tmxContent.append("\t" + key + ":" + sourceSentsMap.get(key)+"\n");        	
        }
        tmxContent.append("\n");
        tmxContent.append("\n"
        		+ "TL sentences per type:\n" ) ;
        for (String key: Lists.newArrayList(targetSentsMap.keySet())) {
            tmxContent.append("\t" + key + ":" + targetSentsMap.get(key)+"\n");        	
        }
        tmxContent.append("\n");

        
        PrintWriter tmxWriter = new PrintWriter(outputTMXList, "UTF-8");
		tmxWriter.print(tmxContent.toString());
		tmxWriter.close();
		logger.info("Created list of tmx results in " + outputTMXList.getAbsolutePath());
	}
	
    public void initialize (String sourceLang, String targetLang) {
    	logger.debug("Inititalizing aligner with " + sourceLang +  " " + targetLang + " " + this.toString());
    	setSourceLang(sourceLang);
    	setTargetLang(targetLang);
    }

	public void initialize(Properties properties) throws Exception {
		this.properties = properties;
    	setSourceLang(properties.getProperty("sourceLang"));
    	setTargetLang(properties.getProperty("targetLang"));
    	if (!properties.isEmpty()) {
        	logger.debug("Inititalizing aligner with properties ");
    		logger.debug(properties.toString());
    	}	
	}
    
	/**
	 * @param file
	 * @param useBoilerplateParagraphs
	 * @return
	 * @throws Exception
	 */
	protected List<String> getParagraphs(File file, boolean useBoilerplateParagraphs) throws Exception {
		List<String> paragraphs = new ArrayList<String>();
		DocumentBuilderFactory dbFactory = DocumentBuilderFactory.newInstance();
		DocumentBuilder dBuilder;
		try {
			dBuilder = dbFactory.newDocumentBuilder();
			Document doc = dBuilder.parse(file);
			NodeList pNodes = doc.getDocumentElement().getElementsByTagName("p");
	        for(int i=0; i< pNodes.getLength(); i++){
	            Element pElement = (Element)pNodes.item(i);;
	            if (pElement.hasAttribute(CRAWLINFO) && pElement.getAttribute(CRAWLINFO).equals(BOILERPLATE) ) {
	            	continue;
	            } else {
	            	paragraphs.add(pElement.getTextContent().trim());
	            }
	        }
		} catch (Exception e) {
			e.printStackTrace();
			throw e;
		}
		return paragraphs;
	}
	/**
	 * @return the sourceLang
	 */
	public String getSourceLang() {
		return sourceLang;
	}
	/**
	 * @param sourceLang the sourceLang to set
	 */
	public void setSourceLang(String sourceLang) {
		this.sourceLang = sourceLang;
	}
	/**
	 * @return the targetLang
	 */
	public String getTargetLang() {
		return targetLang;
	}
	/**
	 * @param targetLang the targetLang to set
	 */
	public void setTargetLang(String targetLang) {
		this.targetLang = targetLang;
	}

	/**
	 * @return the useBoilerplateParagraphs
	 */
	public boolean isUseBoilerplateParagraphs() {
		return useBoilerplateParagraphs;
	}
	/**
	 * @param useBoilerplateParagraphs the useBoilerplateParagraphs to set
	 */
	public void setUseBoilerplateParagraphs(boolean useBoilerplateParagraphs) {
		this.useBoilerplateParagraphs = useBoilerplateParagraphs;
	}

}
